# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import MySQLdb

from sugar_server import env, datastore
from sugar_server.util import enforce


class Cursor(object):

    def __init__(self):
        self._cursor = None

    def __del__(self):
        self.close()

    def insert(self, table_, **kwargs):
        sql = 'INSERT INTO `%s` (' % table_
        for i, field in enumerate(kwargs.keys()):
            if i:
                sql += ', '
            sql += '`%s`' % field
        sql += ') VALUES ('
        for i, value in enumerate(kwargs.values()):
            if i:
                sql += ', '
            sql += _to_str(value)
        sql += ')'
        self._execute(sql)

    def insert_or_update(self, table_, **kwargs):
        try:
            self.insert(table_, **kwargs)
            return True
        except MySQLdb.IntegrityError:
            # Have to avoid using "ON DUPLICATE KEY UPDATE" syntax to make sure
            # that we did exactly UPDATE instead of INSERT
            self.update(table_, **kwargs)
            return False

    def update(self, table_, **kwargs):
        sql = 'UPDATE `%s` SET ' % table_
        where = ''
        for i, (field, value) in enumerate(kwargs.items()):
            if field == datastore.TABLES[table_][0]:
                where = ' WHERE `%s`=%s' % (field, _to_str(value))
            else:
                if i:
                    sql += ', '
                sql += '`%s`=%s' % (field, _to_str(value))
        self._execute(sql + where)

    def select(self, table_, **kwargs):
        sql = 'SELECT * FROM `%s` %s' % (table_, self._compose_query(kwargs))
        self._execute(sql)

        for __ in range(self._cursor.rowcount):
            yield self._cursor.fetchone()

    def delete(self, table_, **kwargs):
        sql = 'DELETE FROM `%s` %s' % (table_, self._compose_query(kwargs))
        self._execute(sql)

    def close(self):
        if self._cursor is not None:
            self._cursor.close()
            self._cursor = None

    def _compose_query(self, kwargs):
        query = ' AND '.join(['`%s`=%s' % (key, _to_str(value)) \
                for key, value in kwargs.items()])
        if query:
            query = 'WHERE ' + query
        return query

    def _execute(self, sql):
        if self._cursor is None:
            self._cursor = _get_cursor()
        try:
            self._cursor.execute(sql)
        except MySQLdb.OperationalError:
            # In case if client lost connection with MySQL server
            self._cursor = _get_cursor(True)
            self._cursor.execute(sql)


def get_db_mtime():
    # TODO
    return None


_connection = None


def _get_cursor(force=False):
    global _connection

    if _connection is None or force:
        format_error = _('MySQL connection string should be in format') + \
                ': user:password@host[:port]'

        enforce(env.db_connect.value, format_error)
        enforce('@' in env.db_connect.value, format_error)
        creds, host = env.db_connect.value.split('@', 1)
        enforce(':' in creds, format_error)
        user, passwd = creds.split(':', 1)
        if ':' in host:
            host, port = host.split(':', 1)
            enforce(port.isdigit(), format_error)
            port = int(port)
        else:
            port = 3306

        _connection = MySQLdb.connect(host=host, port=port,
                user=user, passwd=passwd, db=env.db_name.value)

    return _connection.cursor(MySQLdb.cursors.DictCursor)


def _to_str(value):
    if value is None:
        return "NULL"
    elif isinstance(value, str):
        return '"%s"' % MySQLdb.escape_string(value)
    else:
        return str(value)
