# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import imp
from gettext import gettext as _

from sugar_server import env
from sugar_server.util import enforce


TABLES = {
        # (table_name: (primary_key, not_null, defaults))
        'directory': ('uuid', ['nickname', 'pubkey', 'serial'], {'stolen': 0}),
        }


def select(table_, **kwargs):
    cursor = Cursor()
    try:
        for row in cursor.select(table_, **kwargs):
            yield row
    finally:
        cursor.checkpoint()


def get(table_, **kwargs):
    result = []

    cursor = Cursor()
    try:
        for row in cursor.select(table_, **kwargs):
            result.append(row)
    finally:
        cursor.checkpoint()

    enforce(len(result) <= 1)

    return result[0] if result else {}


def insert(table_, **kwargs):
    cursor = Cursor()
    try:
        cursor.insert(table_, **kwargs)
    finally:
        cursor.checkpoint()


def insert_or_update(table_, **kwargs):
    cursor = Cursor()
    try:
        return cursor.insert_or_update(table_, **kwargs)
    finally:
        cursor.checkpoint()


def delete(table_, **kwargs):
    cursor = Cursor()
    try:
        return cursor.delete(table_, **kwargs)
    finally:
        cursor.checkpoint()


def mtime():
    return Cursor().mtime


class Cursor(object):
    """Low level data access class.

    All changes are assumed not being commited before `checkpoint()` method
    invocation of destroying the `Cursor` object. After invoking
    `checkpoint()`, `Cursor` object might be used to start new transaction.

    """

    _backends = {}

    def __init__(self):
        self._cursor_value = None

        mod_name = env.datastore.value
        if mod_name not in Cursor._backends:
            fp, pathname, desc = imp.find_module(mod_name, [env.BACKENDS_DIR])
            try:
                Cursor._backends[mod_name] = \
                        imp.load_module(mod_name, fp, pathname, desc)
            finally:
                if fp:
                    fp.close()
        self._backend = Cursor._backends[mod_name]

    def __del__(self):
        self.checkpoint()

    @property
    def mtime(self):
        """Timestamp of db modification from previous sugar-server session."""
        return self._backend.get_db_mtime()

    def insert(self, table_, **kwargs):
        """Insert new row.

        :param table_:
            the name of affected table_
        :param kwargs:
            dict of field values; if some of required fields are missed,
            RuntimeError will be raisen

        """
        enforce(table_ in TABLES)
        self._check_insert_integrity(table_, kwargs)
        self._cursor.insert(table_, **kwargs)

    def insert_or_update(self, table_, **kwargs):
        """Insert new or updated existed row.

        :param table_:
            the name of affected table_
        :param kwargs:
            dict of field values; if some of required fields are missed,
            RuntimeError will be raisen
        :returns:
            `True` if new row was added

        """
        enforce(table_ in TABLES)
        self._check_insert_integrity(table_, kwargs)
        return self._cursor.insert_or_update(table_, **kwargs)

    def update(self, table_, **kwargs):
        """Updated existed row.

        :param table_:
            the name of affected table_
        :param kwargs:
            dict of field values; this list should contain primary key as well,
            of cource with current value, to make it possible to find
            the affected row

        """
        enforce(table_ in TABLES)
        self._cursor.update(table_, **kwargs)

    def select(self, table_, **kwargs):
        """Iterate through table rows.

        :param table_:
            the name of affected table_
        :param kwargs:
            dict of field values to concretize the query, might be empty
            to return all rows from the table
        :returns:
            function is an iterator

        """
        enforce(table_ in TABLES)
        for row in self._cursor.select(table_, **kwargs):
            yield row

    def delete(self, table_, **kwargs):
        """Delete existed row.

        :param table_:
            the name of affected table_
        :param kwargs:
            dict of field values to concretize the removing, might be empty
            to delete all rows from the table

        """
        enforce(table_ in TABLES)
        self._cursor.delete(table_, **kwargs)

    def checkpoint(self):
        """Commit all changes made whithin this cursor.

        Any firther operation with the same `Cursor` object will start new
        transaction. Besides, `checkpoint` will be implicitly executed
        on `Cursor` object destroying.

        """
        if self._cursor_value is not None:
            self._cursor_value.close()
            self._cursor_value = None

    @property
    def _cursor(self):
        if self._cursor_value is None:
            self._cursor_value = self._backend.Cursor()
        return self._cursor_value

    def _check_insert_integrity(self, table_, row):
        key, not_null, defaults = TABLES[table_]

        for field, value in defaults.items():
            if field not in row:
                row[field] = value

        enforce(not set(not_null + [key]) - set(row.keys()), \
                _('Datastore integrity error for "%s" table, fields list %r ' \
                'are incomplete to insert new row'), table_, row)
