# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import pwd
import grp
import sys
import imp
import logging
from os.path import join, exists, abspath, dirname, islink, basename
from gettext import gettext as _

from sugar_server import util
from sugar_server.util import enforce


VERSION = '1.0'

# The specs are a little unclear on the encoding of UUIDs, so be
# flexible in what we accept
SERIAL_RE = re.compile('^([A-Z]{3}[A-F0-9]{8})$')
UUID_RE = re.compile('^([a-fA-F0-9-]{32,40})$')

LOGGING_FORMAT = '%(asctime)s %(levelname)s %(name)s: %(message)s'

SERVICES_DIR = join(abspath(dirname(__file__)), 'services')
SERVICES_ALL = sorted([
        _i for _i in os.listdir(SERVICES_DIR) \
                if exists(join(SERVICES_DIR, _i, 'service.py'))])

BACKENDS_DIR = join(abspath(dirname(__file__)), 'backends')
BACKENDS = [_i[:-3] for _i in os.listdir(BACKENDS_DIR) if _i.endswith('.py')]


user = util.Option(
        _('system user and group to start master process'),
        default='sugar_server')

keyring_user = util.Option(
        _('system user to start keyring service'),
        default='sugar_keyring')

debug = util.Option(
        _('debug logging level; multiple argument'),
        default=0, type_cast=int, short_option='-D', action='count')

force = util.Option(
        _('force various operations'),
        default=False, type_cast=bool, short_option='-f', action='store_true')

foreground = util.Option(
        _('Do not send the process into the background'),
        default=False, type_cast=bool, short_option='-F', action='store_true')

root = util.Option(
        _('path to the root directory with server data'),
        default='/library', short_option='-r')

services = util.Option(
        _('list of enabled services, the final list of launched services ' \
                'depends on particular process; empty list means everything'),
        default=set(SERVICES_ALL) - set(['keyring']),
        type_cast=lambda x: set(
                [i for i in re.split('[;,\s]+', x.strip()) if i] \
                        if isinstance(x, str) or isinstance(x, unicode) \
                        else x),
        type_repr=lambda x: ','.join(x), short_option='-s')

hostname = util.Option(
        _('hostname to listen for incomming connections and spreading it ' \
                'as-is among sugar clients'),
        default='schoolserver', short_option='-H')

httpd_port = util.Option(
        _('port for incomming HTTP connections for various services'),
        default=8000, type_cast=int)

datastore = util.Option(
        _('datastore backend type') + ': ' + ' '.join(BACKENDS),
        default='ini')

db_name = util.Option(
        _('database name'),
        default='sugar_server')

db_connect = util.Option(
        _('database connect parameters, if needed, in format') + \
                ': user:password@host[:port]')


def setup(dry_run, logname=None):
    """Setup environment.

    :param dry_run:
        if `True`, do not try to change directories tree
    :param logname:
        if specified, file handler will be created to output logs to the file
        with `logname` filename in standard log directory

    """
    if not debug.value:
        logging_level = logging.WARNING
    elif debug.value == 1:
        logging_level = logging.INFO
    else:
        logging_level = logging.DEBUG

    root_logger = logging.getLogger()
    root_logger.setLevel(logging_level)

    for user_name, mode, path in [
            # root's mode should be 0755 since it is a home directory
            # for sugar-server user and SSH will complain for other modes
            (user.value, 0755, root.value),
            (user.value, 0775, log_path()),
            (user.value, 0775, run_path()),
            (user.value, 0775, var_path()),
            # create directory for keyring service
            (keyring_user.value, 0770, join(root.value, 'keyring')),
            ]:
        if not exists(path):
            enforce(not dry_run, _('Directory %s does not exist'), path)
            logging.info(_('Create %s directory'), path)
            os.makedirs(path)

        stat = os.stat(path)
        uid = pwd.getpwnam(user_name).pw_uid
        gid = grp.getgrnam(user.value).gr_gid

        if uid != stat.st_uid or gid != stat.st_gid:
            enforce(not dry_run, _('Directory %s is not owned by %s'),
                    path, user.value)
            logging.info(_('Change owner for %s directory'), path)
            os.chown(path, uid, gid)

        # We need g+w since there are two users, for master and sign process
        if stat.st_mode & 0777 != mode:
            enforce(not dry_run, _('Directory %s access mode is incorrect'),
                    path)
            os.chmod(path, mode)

    if logname is not False:
        if logname:
            filename = join(log_path(), '%s.log' % logname)
        else:
            filename = None

        for handler in root_logger.handlers:
            root_logger.removeHandler(handler)

        logging.basicConfig(level=logging_level, filename=filename,
                format=LOGGING_FORMAT)


def service(name, dry_run=False):
    """Service's main Python module.

    :param name:
        service name, i.e., sub-directory name in the `services/` directory
    :returns:
        `service.py` module from service sources

    """
    mod, imported = _import(name, 'service', dry_run)
    if imported:
        logging.info(_('Load %s service'), name)
    return mod


def import_from(mod_name, service_name=None):
    """Import python module from service directory.

    Import will happen starting from the caller's directory. It makes sense
    only for importing service modules from service sources, i.e., when module
    paths are not in sys.path to avoid cross-service imports and, thus, module
    names clashes.

    :param mod_name:
        python module name to import
    :param service_name:
        service name, i.e., sub-directory name in the `services/` directory;
        if omitted, caller's directory will be used
    :returns:
        module object

    """
    if not service_name:
        # pylint: disable-msg=W0212
        frame = sys._getframe(1)
        service_name = basename(dirname(frame.f_globals['__file__']))

    return _import(service_name, mod_name)[0]


def get_logger(name, fmt=None):
    """Get service logger.

    :param name:
        name of logger that will be used for log filename as well
    :param fmt:
        if not `None`, new logger will be have its own format
    :returns:
        logging.Logger object

    """
    logger = logging.getLogger(name)
    if not logger.handlers:
        logfile = join(log_path(), '%s.log' % name)
        handler = logging.FileHandler(logfile)
        if fmt is None:
            fmt = LOGGING_FORMAT
        handler.setFormatter(logging.Formatter(fmt))
        logger.addHandler(handler)
    return logger


def assert_serial(serial):
    enforce(SERIAL_RE.match(serial), _('Invalid serial number: %s'), serial)


def assert_uuid(uuid):
    enforce(UUID_RE.match(uuid), _('Invalid UUID: %s'), uuid)


def service_root(*args):
    # pylint: disable-msg=W0212
    frame = sys._getframe(1)
    service_name = basename(dirname(frame.f_globals['__file__']))
    return join(root.value, service_name, *args)


def backup_path(*args):
    return join(root.value, 'backup', *args)


def log_path(*args):
    return join(root.value, 'log', *args)


def run_path(*args):
    return join(root.value, 'run', *args)


def var_path(*args):
    return join(root.value, 'var', *args)


def disk_usage():
    """The storage usage in percents from total numbers.

    :returns:
        the maximum usage between taken bytes and inodes
        on fs where --root is located

    """
    statvfs = _statvfs(root.value)
    if statvfs.f_blocks:
        blocks_used = 100 - statvfs.f_bavail * 100 / statvfs.f_blocks
    else:
        blocks_used = 0
    if statvfs.f_files:
        inodes_used = 100 - statvfs.f_favail * 100 / statvfs.f_files
    else:
        inodes_used = 0
    return max(blocks_used, inodes_used)


def disk_bytes_usage(path=None):
    """How much directory or fs takes in bytes.

    :param path:
        path to the directory to count usage space;
        if `None`, assume entirely fs where --root is located
    :returns:
        size in bytes

    """
    if path is None:
        statvfs = _statvfs(root.value)
        return (statvfs.f_blocks - statvfs.f_bavail) * statvfs.f_bsize
    else:
        return _du(path)


def disk_bytes_total():
    """Whats the storage capacity in bytes.

    :returns:
        size in bytes on the fs where --root is located

    """
    statvfs = _statvfs(root.value)
    return statvfs.f_blocks * statvfs.f_bsize


def disk_bytes_free():
    """How many free space is in bytes.

    :returns:
        free space in bytes on the fs where --root is located

    """
    statvfs = _statvfs(root.value)
    return statvfs.f_bavail * statvfs.f_bsize


def set_statvfs(total, used=0):
    statvfs_file = file(backup_path('.statvfs'), 'w')
    print >> statvfs_file, backup_path()
    print >> statvfs_file, total
    print >> statvfs_file, used
    statvfs_file.close()


def unset_statvfs():
    if exists(backup_path('.statvfs')):
        os.unlink(backup_path('.statvfs'))


def call_crypto(cmd, *args, **kwargs):
    """Run one of bios-crypto utilities.

    :param cmd:
        the name of utility, how it is named in bios-crypto's
        build/ directory in sources tree, to run
    :param args:
        run command arguments
    :param kwargs:
        named arguments to pass to `call` function
    :returns:
        command output, exception otherwise

    """
    cmdline = ['/usr/bin/bc-%s' % cmd] + list([str(i) for i in args])
    return util.assert_call(cmdline, **kwargs)


def _import(name, mod, dry_run=False):
    mod_name = '_%s_%s' % (name, mod)
    result = sys.modules.get(mod_name)
    if result is None and not dry_run:
        mod_path = join(SERVICES_DIR, name)
        try:
            fp, pathname, description = imp.find_module(mod, [mod_path])
            try:
                result = imp.load_module(mod_name, fp, pathname, description)
            finally:
                if fp:
                    fp.close()
        except Exception, error:
            util.exception(_('Fail to import module %s from %s service'),
                    mod, name)
            raise RuntimeError(
                    _('Service "%s" is invalid: %s') % (name, error))
        imported = True
    else:
        imported = False
    return result, imported


def _statvfs(path):
    """Support testing workflow on multi processes level."""
    if not exists(backup_path('.statvfs')):
        return os.statvfs(path)

    class Statvfs(object):

        f_bsize = 1
        f_blocks = 1
        f_bavail = 1
        f_files = 1
        f_favail = 1

    statvfs = Statvfs()
    statvfs_file = file(backup_path('.statvfs'))
    path = statvfs_file.readline().strip()
    statvfs.f_blocks = int(statvfs_file.readline().strip())
    used = int(statvfs_file.readline().strip()) + _du(path, True)
    statvfs.f_bavail = statvfs.f_blocks - used
    return statvfs


def _du(path, force_=False):
    """Support testing workflow on multi processes level."""
    path = abspath(path)

    if not force_ and not exists(backup_path('.statvfs')):
        output = util.assert_call(['/usr/bin/du',
                '--bytes', '--summarize', '--one-file-system', path])
        return int(output.split()[0])

    used = 0
    for root_, dirs, files in os.walk(path):
        for i in dirs[:]:
            if islink(join(root_, i)) or i.startswith('.') and root_ == path:
                dirs.remove(i)
        if root_ != path:
            used += len(files)
    return used


def _populate_services_options_and_commands():

    util.Option.seek('main')
    util.Command.seek('main')
    for i in services.value:
        try:
            util.Option.seek(i, _import(i, 'etc')[0])
            util.Command.seek(i, _import(i, 'etc')[0])
        except ImportError:
            pass


_populate_services_options_and_commands()
