# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time
import fcntl
import errno
import socket
import logging
import SocketServer
from os.path import exists, join
from gettext import gettext as _

from sugar_server import env


_ERROR_MAGIC = 'ERROR:'


class Client(object):
    """Client class for interaction via UNIX socket."""

    def __init__(self, name):
        self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            self._socket.connect(join(env.run_path(), name))
        except Exception:
            logging.error(_('Cannot connect to "%s" IPC server'), name)
            raise

    def send_message(self, message):
        self._socket.send(str(message) + '\n')

    def recv_message(self):
        reply = _recv_line(self._socket)
        if reply.startswith(_ERROR_MAGIC):
            raise RuntimeError(reply.split(':', 1)[-1])
        return reply


class Server(SocketServer.UnixStreamServer, SocketServer.ThreadingMixIn):
    """Server class for interaction via UNIX socket."""

    def __init__(self, name, handle_cb):

        class Handler(SocketServer.BaseRequestHandler):

            def handle(self):
                try:
                    handle_cb(self)
                except Exception, error:
                    logging.exception(
                            _('Cannot process request for %s IPC socket'),
                            name)
                    self.send_message(_ERROR_MAGIC + str(error))

            def send_message(self, message):
                self.request.send(str(message) + '\n')

            def recv_message(self):
                return _recv_line(self.request)

        address = join(env.run_path(), name)
        if exists(address):
            os.unlink(address)

        SocketServer.UnixStreamServer.__init__(self, address, Handler)


def flock(name, timeout=None):
    """Try to exclusively flock the file.

    :param name:
        name of lock file in standard run directory
    :param timeout:
        even if lock is successfully gained it will be ignored
        if file's mtime is less then `timeout` specified in seconds
    :returns:
        `None` if someone locked it already;
        `False` if `timeout` triggered;
        otherwise locking object, call its `checkpoint()` method
        if the process, that asked for locking, was successfully finished

    """
    lock_path = join(env.run_path(), '%s.lock' % name)
    lock = file(lock_path, 'a+')

    try:
        fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError, error:
        lock.close()
        if error.errno in [errno.EACCES, errno.EAGAIN]:
            return None

    lock = _FLock(lock)

    if lock.stale(timeout):
        return lock
    else:
        return False


class _FLock(object):

    def __init__(self, lockfile):
        self._lockfile = lockfile
        self._lockfile.seek(0, 2)

    def __del__(self):
        if self._lockfile is None:
            return
        self._lockfile.close()
        self._lockfile = None

    @property
    def checkpointed(self):
        return self._lockfile.tell()

    def stale(self, timeout):
        shift = int(time.time()) - int(os.stat(self._lockfile.name).st_mtime)
        return not timeout or not self.checkpointed or shift > timeout

    def checkpoint(self):
        if self._lockfile is None:
            return
        self._lockfile.seek(0, 0)
        self._lockfile.write('checkpoint')
        self._lockfile.close()
        self._lockfile = None


def _recv_line(f):
    line = ''
    while True:
        char = f.recv(1)
        if not char or char == '\n':
            break
        line += char
    return line
