# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import datetime
import threading
from os.path import join
from gettext import gettext as _

from sugar_server.util import enforce


def ps():
    """The list of running processes.

    :returns:
        list of dicts with process info, at least pid, cmd and args

    """
    result = []
    for pid in [i for i in os.listdir('/proc') if i.isdigit()]:
        cmdline = file(join('/proc', pid, 'cmdline')).read()
        if not cmdline:
            continue
        entry = {'pid': pid,
                 'cmd': cmdline.split()[0],
                 'args': cmdline,
                 }
        result.append(entry)
    return result


def utcnow():
    """Function only for having a chance to override it in unittest."""
    return datetime.datetime.utcnow()


class ThreadingTCPServer(object):
    """Convenient wrapper around BaseServer to run it in a thread."""

    def __init__(self):
        self._server = None
        self._thread = None

    def start(self, server):
        """Start processing server in a separated thread.

        :param server:
            BaseServer object to run

        """
        self.stop()
        self._server = server
        self._thread = threading.Thread(target=server.serve_forever)
        self._thread.start()

    def stop(self):
        """Stop serving thread.

        Previosly started by `start()` server will be shutting down and serving
        thread joined.

        """
        if self._server is None:
            return
        self._server.shutdown()
        self._thread.join()
        self._server = None
        self._thread = None


class Expiry(object):
    """Represent expiration date."""

    FORMAT = '%Y%m%dT%H%M%SZ'

    # pylint: disable-msg=W0212

    def __init__(self, value=None):
        """Initiate `Expiry` value.

        :param value:
            the initial value, might be `None`, `Expiry`, `int` for days and
            `str` for days or full expiry representation

        """
        is_str = isinstance(value, str) or isinstance(value, unicode)
        if value is None:
            self._abs = None
            self._days = None
        elif isinstance(value, Expiry):
            self._abs = value._abs
            self._days = value._days
        elif isinstance(value, datetime.datetime):
            self._abs = value
            self._days = None
        elif is_str and not value.isdigit():
            if value:
                try:
                    self._abs = datetime.datetime.strptime(
                            value, Expiry.FORMAT)
                except ValueError:
                    raise RuntimeError(_('Absolute expiry "%s", should be ' \
                            'in form of "%s"') % (value, Expiry.FORMAT))
            else:
                self._abs = None
            self._days = None
        elif is_str:
            self._abs = None
            self._days = int(value)
        else:
            enforce(isinstance(value, int), _('Expiry days should be integer'))
            self._abs = None
            self._days = value

    @property
    def days(self):
        """How many days remain for expiration date starting from now."""
        if self._days is not None:
            return self._days
        elif self._abs is not None:
            return max(0, (self._abs - utcnow()).days)

    @property
    def date(self):
        """The exact expiration date starting from now."""
        if self._abs is not None:
            return self._abs
        elif self._days is not None:
            return utcnow() + datetime.timedelta(days=self._days)

    def __str__(self):
        if self.date is None:
            return ''
        else:
            return self.date.strftime(Expiry.FORMAT)

    def __unicode__(self):
        return self.__str__()

    def __repr__(self):
        return self.__str__()

    def __cmp__(self, other):
        return cmp(str(self), str(other))
