# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import signal
import atexit
import logging
from os.path import exists, join, abspath, dirname
from gettext import gettext as _

from sugar_server import env, httpd
from sugar_server.util import enforce


# XXX hack to use os._exit in unittsets to break it immediately
_process_exit = exit


def start(foreground):
    """Start sugar-server process.

    :param foreground:
        do not daemonize sugar-server process
    :returns:
        exit status; `0` on success and `> 0` otherwise

    """
    pidfile, pid = _check_for_instance()
    if pid:
        logging.warning(_('sugar-server is already run with pid %s'), pid)
        return 1

    for i in env.services.value:
        env.service(i).setup()

    if foreground:
        _launch()
    else:
        _daemonize(pidfile, env.log_path('%s.log' % _process_name()), _launch)

    return 0


def stop():
    """Stop sugar-server daemon process.

    :returns:
        exit status; `0` if process was killed and `> 0` otherwise

    """
    __, pid = _check_for_instance()
    if pid:
        os.kill(pid, signal.SIGTERM)
    else:
        logging.warning(_('sugar-server is not run'))
        return 1

    return 0


def started():
    """Is sugar-server process launched in daemon mode."""
    __, pid = _check_for_instance()
    return bool(pid)


def shutdown():
    """Cleanup on exit."""
    for i in env.services.value:
        service = env.service(i, dry_run=True)
        if service is not None:
            service.stop()
    httpd.stop()


def _launch():
    logging.info(_('Start sugar-server'))

    stopped = []

    def sigterm_cb(signum, frame):
        if stopped:
            return
        logging.info(_('Got signal %s, will stop sugar-server'), signum)
        stopped.append(True)

    signal.signal(signal.SIGINT, sigterm_cb)
    signal.signal(signal.SIGTERM, sigterm_cb)

    httpd_services = []

    for i in env.services.value:
        mod = env.service(i)
        mod.start()
        for i in dir(mod):
            if i.startswith('GET_') or i.startswith('POST_'):
                httpd_services.append(mod)
                break

    if httpd_services:
        httpd.start(httpd_services)

    while not stopped:
        signal.pause()


def _process_name():
    if list(env.services.value) == ['keyring']:
        return 'sugar-keyring'
    else:
        return 'sugar-server'


def _check_for_instance():
    pid = None
    pidfile = join(env.run_path(), '%s.pid' % _process_name())

    if exists(pidfile):
        try:
            pid = int(file(pidfile).read().strip())
            os.getpgid(pid)
        except (ValueError, OSError):
            pid = None

    return pidfile, pid


def _daemonize(pid_path, log_path, doer):
    """Detach a process from the controlling terminal and run in background."""

    pid_path = abspath(pid_path)
    enforce(os.access(dirname(pid_path), os.W_OK))
    log_path = abspath(log_path)
    enforce(os.access(dirname(log_path), os.W_OK))

    if os.fork() > 0:
        # Exit parent of the first child
        return

    # Decouple from parent environment
    os.chdir(os.sep)
    os.setsid()

    if os.fork() > 0:
        # Exit from second parent
        # pylint: disable-msg=W0212
        os._exit(0)

    # Redirect standard file descriptors
    sys.stdout.flush()
    sys.stderr.flush()
    stdin = file('/dev/null')
    logfile = file(log_path, 'a+')
    os.dup2(stdin.fileno(), sys.stdin.fileno())
    os.dup2(logfile.fileno(), sys.stdout.fileno())
    os.dup2(logfile.fileno(), sys.stderr.fileno())

    pidfile = file(pid_path, 'w')
    pidfile.write(str(os.getpid()))
    pidfile.close()
    atexit.register(lambda: os.remove(pid_path))

    try:
        doer()
        status = 0
    except Exception:
        logging.exception(_('Abort sugar-server due to error'))
        status = 1

    shutdown()
    _process_exit(status)
