# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
from os.path import exists
from gettext import gettext as _

from sugar_server import env, datastore, ipc, misc
from sugar_server.util import enforce


etc = env.import_from('etc')

_logger = env.get_logger('activation')


class Lease(object):

    def __init__(self, content, serial=None, delegated=False):
        """Parse lease content.

        :param content:
            lease content
        :param serial:
            if specified, compare it with serial number parsed from
            lease content to make sure if lease is valid

        """
        self._content = content.strip()
        parts = self.content.split(' ', 4)
        enforce(len(parts) == 5 and parts[0] == 'act01:' and parts[2] == 'K',
                _('Incorrect lease "%s"'), self.content)
        self._serial = parts[1]
        self._expiry = misc.Expiry(parts[3])
        if serial:
            enforce(self.serial == serial,
                    _('Found lease is not for %s'), serial)
        self._row = None
        self._delegated = delegated

    @property
    def content(self):
        return self._content

    @property
    def serial(self):
        return self._serial

    @property
    def uuid(self):
        return self.fields['uuid']

    @property
    def expiry(self):
        return self._expiry

    @property
    def expired(self):
        return self._expiry.date < misc.utcnow()

    @property
    def fields(self):
        if self._row is None:
            matched = [i for i in \
                    datastore.select('directory', serial=self.serial)]
            enforce(matched, KeyError, _('Unknown serial %s'), self.serial)
            self._row = matched[0]
        return self._row

    @property
    def stolen(self):
        try:
            return self.fields['stolen'] != '0'
        except KeyError:
            pass

    @property
    def delegated(self):
        return self._delegated

    def get_stolen_status(self, nonce):
        if self.stolen:
            payload = '%s:%s:STOLEN' % (self.fields['uuid'], nonce)
        else:
            payload = '%s:%s' % (self.fields['uuid'], nonce)
        return hashlib.sha256(payload).hexdigest()


class Delegation(object):

    def __init__(self, content, serial=None):
        """Parse delegation content.

        :param content:
            delegation content
        :param serial:
            if specified, compare it with serial number parsed from
            lease content to make sure if lease is valid

        """
        self._content = content.strip()
        parts = self.content.split(' ', 4)
        enforce(len(parts) == 5 and parts[0] == 'del01:' and \
                parts[3] == 'sig02:', _('Incorrect delegation "%s"'),
                self.content)
        self._serial = parts[1]
        if serial:
            enforce(self.serial == serial,
                    _('Found delegation is not for %s'), serial)
        self._chain = ' '.join(parts[-2:])

    @property
    def content(self):
        return self._content

    @property
    def serial(self):
        return self._serial

    @property
    def chain(self):
        return self._chain


def get(serial, expiry=None):
    lease_path = etc.lease_path(serial)
    if exists(lease_path):
        try:
            lease = Lease(file(lease_path).read(), serial)
            if lease.expired:
                _logger.info(_('Pre-existed lease expiried for %s, ' \
                        'will delegate'), serial)
            else:
                _logger.info(_('Found Pre-existed lease for %s'), serial)
                return lease
        except Exception:
            _logger.exception(_('Incorrect pre-existed lease, ' \
                    'will try to gelegate new one'))
    else:
        enforce(etc.delegation.value,
                _('No pre-existed lease and delegation is disabled'))
        _logger.info(_('No pre-existed lease for %s, will delegate'), serial)

    if expiry is None:
        _logger.info(_('Use default lease expiry for %s, %s'),
                serial, etc.default_expiry.value)
        expiry = etc.default_expiry.value

    lease = delegate(serial, expiry)
    return Lease(lease, serial, delegated=True)


def sign_time(serial, nonce, expiry=1):
    # XXX hardcoded in xs-activation/oat.py sources
    now = '00110529T135103Z'
    payload = '%s:%s' % (now, nonce)
    signed = sign(serial, expiry, payload)
    return 'time01: %s %s' % (now, signed)


def sign(serial, expiry, data):
    oats_path = etc.oats_delegations_path(serial)
    enforce(oats_path, _('No OAT delegation for %s'), serial)
    oats_delegation = Delegation(file(oats_path).read(), serial)

    args = {'serial': serial,
            'expiry': misc.Expiry(expiry),
            'data': data,
            }
    payload = '%(serial)s:%(expiry)s:%(data)s' % args
    signed = _call_keyring('sign', expiry, payload)
    _sig02_prefix_to_skip_, signed = signed.split(' ', 1)

    return '%s %s' % (oats_delegation.chain, signed)


def delegate(serial, expiry):
    expiry = misc.Expiry(expiry)

    ld_path = etc.lease_delegations_path(serial)
    enforce(exists(ld_path), _('No lease delegation for %s'), serial)

    ld_fields = file(ld_path).read().split(' ', 3)
    enforce(len(ld_fields) == 4 and ld_fields[0] == 'del01:',
            _('Incorrect delegation header in %s for %s'), ld_path, serial)

    ld_serial, uuid, chain = ld_fields[1:]
    enforce(ld_serial == serial,
            _('Wrong serial number in %s, it should be %s'), ld_path, serial)

    if etc.max_expiry.value is not None and expiry > etc.max_expiry.value:
        _logger.info(_('Specified expiry for %s, %s, is too long, ' \
                'will use the possible maximum, %s'),
                serial, expiry, etc.max_expiry.value)
        expiry = etc.max_expiry.value

    args = {'serial': serial,
            'uuid': uuid,
            'expiry': str(expiry),
            'chain': chain.split('\n')[0],
            }
    prefix = 'act01: %(serial)s K %(expiry)s %(chain)s' % args
    payload = '%(serial)s:%(expiry)s:%(serial)s:%(uuid)s:K:%(expiry)s' % args

    return prefix + _call_keyring('sign', expiry, payload)[6:]


def _call_keyring(*args):
    client = ipc.Client('keyring')
    for i in args:
        client.send_message(i)
    return client.recv_message()
