# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys
import logging
import SocketServer
from os.path import join, exists, dirname, isdir
from gettext import gettext as _

from sugar_server import env, util, misc
from sugar_server.util import enforce


etc = env.import_from('etc')
leases = env.import_from('leases')

_MAX_REQUEST_SIZE = 512
_COMMAND_RE = re.compile('^([a-zA-Z0-9]+): (.*)$')

_logger = env.get_logger('activation')
_service = misc.ThreadingTCPServer()


def setup():
    for path in ['lease-delegations', 'oats-delegations', 'leases']:
        path = env.service_root(path)
        if not exists(path):
            logging.info(_('Create %s directory'), path)
            os.makedirs(path)


def start():
    _service.start(_Server())


def stop():
    _service.stop()


def cmd_import_dir(args):
    enforce(args, _('Command "import" requires PATH argument'))
    path = args.pop(0)
    enforce(isdir(path), _('Path %s is not a dir'), path)

    for filename in os.listdir(path):
        filepath = join(path, filename)
        if filename in ['lease.sig', 'leases.dat']:
            cmd_import_lease([filepath])
        elif filename == 'd-lease.sig':
            cmd_import_lease_delegation([filepath])
        elif filename == 'd-oats.sig':
            cmd_import_oats_delegation([filepath])
        else:
            _logger.info(_('Unrecognized file %s'), filepath)


def cmd_import_lease(args):
    _import(_('lease'), args, etc.lease_path,
            lambda content: leases.Lease(content).serial)


def cmd_import_lease_delegation(args):
    _import(_('lease delegation'), args, etc.lease_delegations_path,
            lambda content: leases.Delegation(content).serial)


def cmd_import_oats_delegation(args):
    _import(_('OATS delegation'), args, etc.oats_delegations_path,
            lambda content: leases.Delegation(content).serial)


class Handler(SocketServer.BaseRequestHandler):

    def handle(self):
        # The request cannot be bigger
        request = self.request.recv(_MAX_REQUEST_SIZE)
        try:
            self.do_handle(request)
        except Exception:
            _logger.exception(_('Cannot process request "%s"'), request)
            self.request.send('UNKNOWN\n')

    def do_handle(self, request):
        match = env.SERIAL_RE.match(request)
        if match is not None:
            _process_lease_request(self.request, match.group(1))
        else:
            match = _COMMAND_RE.match(request)
            if match is not None:
                command, args = match.groups()
                if command == 'time01':
                    _process_time_request(self.request, args)
                else:
                    raise RuntimeError(_('Unknown command %s') % command)
            else:
                raise RuntimeError(_('Cannot parse request "%s"') % request)


class _Server(SocketServer.TCPServer, SocketServer.ThreadingMixIn):

    def __init__(self):
        SocketServer.TCPServer.__init__(self,
                (env.hostname.value, etc.activation_port.value),
                Handler)
        self.allow_reuse_address = True


def _process_lease_request(request, serial):
    lease = leases.get(serial)
    if lease.stolen:
        _logger.info(_('Requested lease for %s is stolen'), serial)
        request.send('STOLEN')
    else:
        _logger.info(_('Return lease for %s'), serial)
        request.send(lease.content)


def _process_time_request(request, args_str):
    args = args_str.split()
    enforce(len(args) == 2, _('Wrong number of arguments for time01 command'))

    serial, nonce = args
    env.assert_serial(serial)
    enforce(len(nonce) == 22,
            _('time01 command argument "nonce", should be 22 chars long'))

    reply = leases.sign_time(serial, nonce)
    request.send(reply)


def _import(title, path, get_path_cb, get_serial_cb):

    def write(line):
        try:
            line = line.strip() + '\n'
            serial = get_serial_cb(line)
            dst_path = get_path_cb(serial)

            _logger.info(_('Import %s for serial %s'), title, serial)

            if not exists(dirname(dst_path)):
                os.makedirs(dirname(dst_path))
            f = file(dst_path, 'w')
            f.write(line)
            f.close()
        except Exception:
            _logger.exception(_('Cannot import %s "%s" from %s'),
                    title, line, path)

    if path:
        path = path.pop(0)
        fin = file(path)
    else:
        path = 'stdin'
        fin = sys.stdin
    try:
        json_text = ''

        for line in fin:
            line = line.strip()
            if not line:
                continue

            if json_text is not None and (json_text or line.startswith('[')):
                json_text += line
            else:
                json_text = None
                try:
                    write(line)
                except Exception:
                    _logger.exception(_('Cannot import %s "%s" from %s'),
                            title, line, path)

        if json_text:
            try:
                # It seems that there is no need in checking `_version` value
                # since `contnent` entries have version info anyway
                _version, content = util.json.loads(json_text)
                for i in content.values() \
                        if isinstance(content, dict) else content:
                    write(i)
            except Exception:
                _logger.exception(_('Cannot import %s from %s'), title, path)
    finally:
        if fin is not sys.stdin:
            fin.close()
