# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pwd
import threading
from ConfigParser import ConfigParser
from os.path import join, exists, basename, abspath
from gettext import gettext as _

from sugar_server import env, datastore, misc, printf, util
from sugar_server.util import enforce


etc = env.import_from('etc')
trimmer = env.import_from('trimmer')
rsync = env.import_from('rsync')


_LOADAVG_MAX = 5
_CONCURRENT_RSYNCS_MAX = 10

_logger = env.get_logger('backup')

_service_cond = threading.Condition()
_service_exit = False


def setup():
    enforce(pwd.getpwnam(env.user.value).pw_dir == env.root.value,
            _('Home directory for %s user should be %s to let users ' \
                    'do backups via SSH'), env.user, env.root)

    if not exists(env.backup_path()):
        _logger.info(_('Create %s backup directory'), env.backup_path())
        os.makedirs(env.backup_path())
    if os.stat(env.backup_path()).st_mode & 0777 != 0700:
        _logger.info(_('Change access mode for %s backup directory'),
                env.backup_path())
        os.chmod(env.backup_path(), 0700)

    rsync_src = join(env.SERVICES_DIR, 'backup', 'sugar-server-rsync')
    rsync_dst = env.backup_path('sugar-server-rsync')
    rsync_src_stat = os.stat(rsync_src)

    if not exists(rsync_dst) or \
            rsync_src_stat.st_mtime != os.stat(rsync_dst).st_mtime:
        _logger.info(_('Update sugar-server-rsync rsync wrapper script'))

        rsync_txt = []
        for line in file(rsync_src).readlines():
            pypath = abspath(join(env.SERVICES_DIR, '..', '..'))
            rsync_txt.append(line.replace('@SUGAR_SERVER_PYTHONPATH@', pypath))

        rsync_file = file(rsync_dst, 'w')
        rsync_file.write(''.join(rsync_txt))
        rsync_file.close()
        os.chmod(rsync_dst, 0755)
        os.utime(rsync_dst, (rsync_src_stat.st_mtime, rsync_src_stat.st_mtime))

    rsync_conf = ConfigParser()
    for prop in util.Option.items.values():
        if not rsync_conf.has_section(prop.section):
            rsync_conf.add_section(prop.section)
        rsync_conf.set(prop.section, prop.name, str(prop))
    rsync_conf_file = file(env.backup_path('sugar-server-rsync.conf'), 'w')
    rsync_conf.write(rsync_conf_file)
    rsync_conf_file.close()


def start():

    def service():
        _service_cond.acquire()
        while True:
            trimmer.trim()
            _service_cond.wait(etc.trim_timeout.value)
            if _service_exit:
                break
        _service_cond.release()

    threading.Thread(target=service).start()


def stop():
    _wakeup(True)


def GET_backup(request):
    path_parts = request.path.split('/')
    enforce(len(path_parts) == 5, _('Incorrect url path to backup service'))
    __, __, proto_version, magic_word, serial = path_parts
    enforce(proto_version == '1', _('Incorrect backup protocol version'))
    enforce(magic_word == 'available',
            _('Incorrect backup availability request'))

    uuid = datastore.get('directory', serial=serial).get('uuid')
    if uuid is None:
        request.reply(404)
        return

    backup_path = join(env.backup_path(), uuid)
    enforce(exists(backup_path),
            _('Serial %s is not registered for backups'), serial)

    if flow_opened():
        request.reply(200)
    else:
        request.reply(503)


def flow_opened():
    loadavg = os.getloadavg()[0]
    if loadavg > _LOADAVG_MAX:
        _logger.info(_('The system load avarage is too high, %s > %s. ' \
                'Will deny it'), loadavg, _LOADAVG_MAX)
        return False

    if env.disk_usage() >= etc.hard_quota.value:
        _logger.warning(_('No free storage space to accept new backup, ' \
                '%s >= %s'), env.disk_usage(), etc.hard_quota.value)
        _wakeup()
        return False

    concurrent_rsyncs = \
            len([i for i in misc.ps() if basename(i['cmd']) == 'rsync'])
    if concurrent_rsyncs > _CONCURRENT_RSYNCS_MAX:
        _logger.info(_('Too many rsync concurrent sessions, %s > %s. ' \
                'Will deny backup'), concurrent_rsyncs, _CONCURRENT_RSYNCS_MAX)
        return False

    return True


def cmd_trim(args):
    result = trimmer.trim(env.force.value)

    if result is trimmer.NO_NEED:
        printf.info(_('No need in trimming at all'))
    elif result is trimmer.ONGOING:
        printf.info(_('Trimming is already started by someone else'))
    elif result is trimmer.SKIPPED:
        printf.info(_('Regular trimming already finished, no need ' \
                'this time. Pass --force argument to force trimming'))
    elif result is trimmer.SUCCESS:
        printf.info(_('Trimming finished successfully'))


def _wakeup(is_exit=None):
    global _service_exit
    _service_cond.acquire()
    _service_exit = is_exit
    _service_cond.notify()
    _service_cond.release()
