# Copyright (C) 2011, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pwd
import grp
from os.path import join, exists, dirname
from gettext import gettext as _

from sugar_server import env, util, datastore
from sugar_server.util import enforce


_logger = env.get_logger('id')
_directory = None
_authorized_keys_path = join(env.root.value, '.ssh', 'authorized_keys')


def setup():
    backup_ssh_path = dirname(_authorized_keys_path)
    if not exists(backup_ssh_path):
        _logger.info(_('Create %s SSH directory for backup user'),
                backup_ssh_path)
        os.makedirs(backup_ssh_path)

    stat = os.stat(backup_ssh_path)
    if stat.st_mode & 0777 != 0700:
        _logger.info(_('Change access mode of %s SSH directory ' \
                'for backup user'), backup_ssh_path)
        os.chmod(backup_ssh_path, 0700)

    uid = pwd.getpwnam(env.user.value).pw_uid
    gid = grp.getgrnam(env.user.value).gr_gid
    enforce(stat.st_uid == uid and stat.st_gid == gid,
            _('Wrong user for %s, it should be %s'), backup_ssh_path, env.user)

    if not exists(_authorized_keys_path):
        _create_keys()
    else:
        stat = os.stat(_authorized_keys_path)
        if stat.st_mode & 0777 != 0600 or \
                stat.st_uid != uid or stat.st_gid != gid or \
                stat.st_mtime != datastore.mtime():
            _create_keys()


def close():
    if exists(_authorized_keys_path):
        mtime = int(datastore.mtime() or 0)
        os.utime(_authorized_keys_path, (mtime, mtime))


def register(uuid, nickname, pubkey, serial):
    env.assert_serial(serial)
    env.assert_uuid(uuid)
    enforce('\n' not in nickname, _('Invalid nickname: %s'), nickname)
    # TODO should precise match pubkeys
    enforce('\n' not in pubkey, _('Invalid public key: %s'), pubkey)

    # sugar-toolkit returns keys w/o prefix
    pubkey = 'ssh-dss %s' % pubkey

    inserted = datastore.insert_or_update('directory',
            uuid=uuid, nickname=nickname, pubkey=pubkey, serial=serial)

    if inserted:
        _logger.info('Registered new student')
    else:
        _logger.info('Changed registeration for existed student')

    if inserted and exists(_authorized_keys_path):
        keys_file = file(_authorized_keys_path, 'a')
        _add_authorized_key(keys_file, uuid, pubkey)
        keys_file.close()
    else:
        # registeration was overriden, just recreate authorized_keys
        # instead of changing its content correspondingly
        _create_keys()

    backup_dir = join(env.backup_path(), uuid)
    if not exists(backup_dir):
        os.makedirs(backup_dir)


def _create_keys():
    _logger.debug('Create authorized_keys file')

    keys_file = util.new_file(_authorized_keys_path, 0600)
    for row in datastore.select('directory'):
        _add_authorized_key(keys_file, row['uuid'], row['pubkey'])
    keys_file.close()


def _add_authorized_key(fd, uuid, pubkey):
    fd.write('command="%s %s",no-port-forwarding,' \
             'no-X11-forwarding,no-agent-forwarding,no-pty %s\n' % \
             (env.backup_path('sugar-server-rsync'), uuid, pubkey))
